"""Find extra requirements."""

import argparse
import collections
import importlib.metadata
import logging
import os
import sys
from pathlib import Path
from typing import Callable, Iterable, List, Optional, Union
from unittest import mock

from packaging.utils import canonicalize_name
from pip._internal.commands.show import search_packages_info
from pip._internal.req.req_file import ParsedRequirement

from pip_check_reqs import common
from pip_check_reqs.common import version_info

log = logging.getLogger(__name__)


def find_extra_reqs(
    requirements_filename: Path,
    paths: Iterable[Path],
    ignore_files_function: Callable[[str], bool],
    ignore_modules_function: Callable[[str], bool],
    ignore_requirements_function: Callable[
        [Union[str, ParsedRequirement]], bool
    ],
    skip_incompatible: bool,
) -> List[str]:
    # 1. find files used by imports in the code (as best we can without
    #    executing)
    used_modules = common.find_imported_modules(
        paths=paths,
        ignore_files_function=ignore_files_function,
        ignore_modules_function=ignore_modules_function,
    )

    # 2. find which packages provide which files
    installed_files = {}
    all_pkgs = [
        dist.metadata["Name"] for dist in importlib.metadata.distributions()
    ]

    # On Python 3.11 (and maybe higher), setting this environment variable
    # dramatically improves speeds.
    # See https://github.com/r1chardj0n3s/pip-check-reqs/issues/123.
    with mock.patch.dict(os.environ, {"_PIP_USE_IMPORTLIB_METADATA": "False"}):
        packages_info = list(search_packages_info(all_pkgs))

    for package in packages_info:
        package_name = package.name
        package_location = package.location
        package_files = []
        for item in package.files or []:
            here = Path(".").resolve()
            item_location_rel = Path(package_location) / item
            item_location = item_location_rel.resolve()
            try:
                relative_item_location = item_location.relative_to(here)
            except ValueError:
                # Ideally we would use Pathlib.is_relative_to rather than
                # checking for a ValueError, but that is only available in
                # Python 3.9+.
                relative_item_location = item_location
            package_files.append(str(relative_item_location))

        log.debug(
            "installed package: %s (at %s)", package_name, package_location
        )
        for package_file in package_files:
            path = os.path.realpath(
                os.path.join(package_location, package_file),
            )
            installed_files[path] = package_name
            package_path = common.is_package_file(path)
            if package_path:
                # we've seen a package file so add the bare package directory
                # to the installed list as well as we might want to look up
                # a package by its directory path later
                installed_files[package_path] = package_name

    # 3. match imported modules against those packages
    used = collections.defaultdict(list)

    for modname, info in used_modules.items():
        # probably standard library if it's not in the files list
        if info.filename in installed_files:
            used_name = canonicalize_name(installed_files[info.filename])
            log.debug(
                "used module: %s (from package %s)",
                modname,
                installed_files[info.filename],
            )
            used[used_name].append(info)
        else:
            log.debug(
                "used module: %s (from file %s, assuming stdlib or local)",
                modname,
                info.filename,
            )

    # 4. compare with requirements
    explicit = common.find_required_modules(
        ignore_requirements_function=ignore_requirements_function,
        skip_incompatible=skip_incompatible,
        requirements_filename=requirements_filename,
    )

    return [name for name in explicit if name not in used]


def main(arguments: Optional[List[str]] = None) -> None:
    """Main entry point."""
    usage = "usage: %prog [options] files or directories"
    parser = argparse.ArgumentParser(usage)
    parser.add_argument("paths", type=Path, nargs="*")
    parser.add_argument(
        "--requirements-file",
        dest="requirements_filename",
        type=Path,
        metavar="PATH",
        default=Path("requirements.txt"),
        help="path to the requirements file "
        '(defaults to "requirements.txt")',
    )
    parser.add_argument(
        "-f",
        "--ignore-file",
        dest="ignore_files",
        action="append",
        default=[],
        help="file paths globs to ignore",
    )
    parser.add_argument(
        "-m",
        "--ignore-module",
        dest="ignore_mods",
        action="append",
        default=[],
        help="used module names (globs are ok) to ignore",
    )
    parser.add_argument(
        "-r",
        "--ignore-requirement",
        dest="ignore_reqs",
        action="append",
        default=[],
        help="reqs in requirements to ignore",
    )
    parser.add_argument(
        "-s",
        "--skip-incompatible",
        dest="skip_incompatible",
        action="store_true",
        default=False,
        help="skip requirements that have incompatible " "environment markers",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        default=False,
        help="be more verbose",
    )
    parser.add_argument(
        "-d",
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="be *really* verbose",
    )
    parser.add_argument(
        "-V",
        "--version",
        dest="version",
        action="store_true",
        default=False,
        help="display version information",
    )

    parse_result = parser.parse_args(arguments)

    if parse_result.version:
        print(version_info())
        sys.exit(0)

    if not parse_result.paths:
        parser.error("no source files or directories specified")

    ignore_files = common.ignorer(parse_result.ignore_files)
    ignore_mods = common.ignorer(parse_result.ignore_mods)
    ignore_reqs = common.ignorer(parse_result.ignore_reqs)

    logging.basicConfig(format="%(message)s")
    if parse_result.debug:
        level = logging.DEBUG
    elif parse_result.verbose:
        level = logging.INFO
    else:
        level = logging.WARN
    log.setLevel(level)
    common.log.setLevel(level)

    log.info(version_info())

    extras = find_extra_reqs(
        requirements_filename=parse_result.requirements_filename,
        paths=parse_result.paths,
        ignore_files_function=ignore_files,
        ignore_modules_function=ignore_mods,
        ignore_requirements_function=ignore_reqs,
        skip_incompatible=parse_result.skip_incompatible,
    )

    if extras:
        log.warning("Extra requirements:")
    for name in extras:
        message = f"{name} in {parse_result.requirements_filename}"
        log.warning(message)

    if extras:
        sys.exit(1)



"""Common functions."""

import ast
import fnmatch
import imp
import logging
import os
import re
import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import (
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

from packaging.markers import Marker
from packaging.utils import NormalizedName, canonicalize_name
from pip._internal.network.session import PipSession
from pip._internal.req.constructors import install_req_from_line
from pip._internal.req.req_file import ParsedRequirement, parse_requirements

from . import __version__

log = logging.getLogger(__name__)


@dataclass
class FoundModule:
    """A module with uses in the source."""

    modname: str
    filename: str
    locations: List[Tuple[str, int]] = field(default_factory=list)

    def __post_init__(self) -> None:
        self.filename = os.path.realpath(self.filename)


class _ImportVisitor(ast.NodeVisitor):
    def __init__(self, ignore_modules_function: Callable[[str], bool]) -> None:
        super().__init__()
        self._ignore_modules_function = ignore_modules_function
        self._modules: Dict[str, FoundModule] = {}
        self._location: Optional[str] = None

    def set_location(self, location: str) -> None:
        self._location = location

    # Ignore the name error as we are overriding the method.
    def visit_Import(  # pylint: disable=invalid-name
        self,
        node: ast.Import,
    ) -> None:
        for alias in node.names:
            self._add_module(alias.name, node.lineno)

    # Ignore the name error as we are overriding the method.
    def visit_ImportFrom(  # pylint: disable=invalid-name
        self,
        node: ast.ImportFrom,
    ) -> None:
        if node.module == "__future__":
            # not an actual module
            return
        for alias in node.names:
            if node.module is None or node.level != 0:
                # relative import
                continue
            self._add_module(node.module + "." + alias.name, node.lineno)

    def _add_module(self, modname: str, lineno: int) -> None:
        if self._ignore_modules_function(modname):
            return
        path = None
        progress = []
        modpath = last_modpath = None
        for modname_part in modname.split("."):
            try:
                _, modpath, _ = imp.find_module(modname_part, path)
            except ImportError:
                # the component specified at this point is not importable
                # (is just an attribute of the module)
                # *or* it's not actually installed, so we don't care either
                break

            # success! we found *something*
            progress.append(modname_part)

            # we might have previously seen a useful path though...
            if modpath is None:
                # the `sys` module will hit this code path, and `os` will on
                # 3.11+.
                # Possibly others will, but I've not discovered them.
                modpath = last_modpath
                break

            # ... though it might not be a file, so not interesting to us
            if not os.path.isdir(modpath):
                break

            path = [modpath]
            last_modpath = modpath

        if modpath is None:
            # the module doesn't actually appear to exist on disk
            return

        modname = ".".join(progress)
        if modname not in self._modules:
            self._modules[modname] = FoundModule(modname, modpath)
        assert isinstance(self._location, str)
        self._modules[modname].locations.append((self._location, lineno))

    def finalise(self) -> Dict[str, FoundModule]:
        result = self._modules
        return result


def pyfiles(root: Path) -> Generator[Path, None, None]:
    if root.is_file():
        if root.suffix == ".py":
            yield root.absolute()
        else:
            raise ValueError(f"{root} is not a python file or directory")
    elif root.is_dir():
        for item in root.rglob("*.py"):
            yield item.absolute()


def find_imported_modules(
    paths: Iterable[Path],
    ignore_files_function: Callable[[str], bool],
    ignore_modules_function: Callable[[str], bool],
) -> Dict[str, FoundModule]:
    vis = _ImportVisitor(ignore_modules_function=ignore_modules_function)
    for path in paths:
        for filename in pyfiles(path):
            if ignore_files_function(str(filename)):
                log.info("ignoring: %s", os.path.relpath(filename))
                continue
            log.debug("scanning: %s", os.path.relpath(filename))
            with open(filename, encoding="utf-8") as file_obj:
                content = file_obj.read()
            vis.set_location(str(filename))
            vis.visit(ast.parse(content, str(filename)))
    return vis.finalise()


def find_required_modules(
    ignore_requirements_function: Callable[
        [Union[str, ParsedRequirement]], bool
    ],
    skip_incompatible: bool,
    requirements_filename: Path,
) -> Set[NormalizedName]:
    explicit = set()
    for requirement in parse_requirements(
        str(requirements_filename), session=PipSession()
    ):
        requirement_name = install_req_from_line(
            requirement.requirement,
        ).name
        assert isinstance(requirement_name, str)

        if ignore_requirements_function(requirement):
            log.debug("ignoring requirement: %s", requirement_name)
            continue

        if skip_incompatible:
            requirement_string = requirement.requirement
            if not has_compatible_markers(requirement_string):
                log.debug(
                    "ignoring requirement (incompatible environment "
                    "marker): %s",
                    requirement_string,
                )
                continue

        log.debug("found requirement: %s", requirement_name)
        explicit.add(canonicalize_name(requirement_name))

    return explicit


def has_compatible_markers(full_requirement: str) -> bool:
    if ";" not in full_requirement:
        return True  # No environment marker.

    enviroment_marker = full_requirement.split(";")[1]
    if not enviroment_marker:
        return True  # Empty environment marker.

    return Marker(enviroment_marker).evaluate()


def is_package_file(path: str) -> str:
    """Determines whether the path points to a Python package sentinel
    file - the __init__.py or its compiled variants.
    """
    search_result = re.search(r"(.+)/__init__\.py[co]?$", path)
    if search_result is not None:
        return search_result.group(1)
    return ""


def ignorer(ignore_cfg: List[str]) -> Callable[..., bool]:
    if not ignore_cfg:
        return lambda candidate: False

    def ignorer_function(
        candidate: Union[str, ParsedRequirement],
        ignore_cfg: List[str] = ignore_cfg,
    ) -> bool:
        for ignore in ignore_cfg:
            if isinstance(candidate, str):
                candidate_path = candidate
            else:
                optional_candidate_path = install_req_from_line(
                    candidate.requirement,
                ).name
                assert isinstance(optional_candidate_path, str)
                candidate_path = optional_candidate_path

            if fnmatch.fnmatch(candidate_path, ignore):
                return True
            if fnmatch.fnmatch(os.path.relpath(candidate_path), ignore):
                return True
        return False

    return ignorer_function


def version_info() -> str:
    major, minor, patch = sys.version_info[:3]
    python_version = f"{major}.{minor}.{patch}"
    parent_directory = Path(__file__).parent.resolve()
    return (
        f"pip-check-reqs {__version__} "
        f"from {parent_directory} "
        f"(python {python_version})"
    )



from codecs import open
from os import path
from pathlib import Path
from typing import List

from setuptools import setup

from pip_check_reqs import __version__

here = path.abspath(path.dirname(__file__))


def _get_dependencies(requirements_file: Path) -> List[str]:
    """
    Return requirements from a requirements file.
    This expects a requirements file with no ``--find-links`` lines.
    """
    lines = requirements_file.read_text().strip().split("\n")
    return [line for line in lines if not line.startswith("#")]


with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

with open(path.join(here, "CHANGELOG.rst"), encoding="utf-8") as f:
    long_description += f.read()

INSTALL_REQUIRES = _get_dependencies(
    requirements_file=Path("requirements.txt"),
)

DEV_REQUIRES = _get_dependencies(
    requirements_file=Path("test-requirements.txt"),
)

setup(
    name="pip_check_reqs",
    version=__version__,
    description=(
        "Find packages that should or should not be in requirements for a "
        "project"
    ),
    long_description=long_description,
    url="https://github.com/r1chardj0n3s/pip-check-reqs",
    author="Richard Jones",
    author_email="r1chardj0n3s@gmail.com",
    maintainer="Adam Dangoor",
    maintainer_email="adamdangoor@gmail.com",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    python_requires=">=3.8.0",
    packages=["pip_check_reqs"],
    entry_points={
        "console_scripts": [
            "pip-missing-reqs=pip_check_reqs.find_missing_reqs:main",
            "pip-extra-reqs=pip_check_reqs.find_extra_reqs:main",
        ],
    },
    install_requires=INSTALL_REQUIRES,
    extras_require={"dev": DEV_REQUIRES},
)
